# -*- coding: utf-8 -*-
# flake8: noqa
"""
hyper/compat
~~~~~~~~~~~~

Normalizes the Python 2/3 API for internal use.
"""
from contextlib import contextmanager
import sys
import zlib

try:
    from . import ssl_compat
except ImportError:
    # TODO log?
    ssl_compat = None

_ver = sys.version_info
is_py2 = _ver[0] == 2
is_py2_7_9_or_later = _ver[0] >= 2 and _ver[1] >= 7 and _ver[2] >= 9
is_py3 = _ver[0] == 3
is_py3_3 = is_py3 and _ver[1] == 3


@contextmanager
def ignore_missing():
    try:
        yield
    except (AttributeError, NotImplementedError):  # pragma: no cover
        pass

if is_py2:
    if is_py2_7_9_or_later:
        import ssl
    else:
        ssl = ssl_compat

    from urllib import urlencode
    from urlparse import urlparse, urlsplit
    from itertools import imap

    def to_byte(char):
        return ord(char)

    def decode_hex(b):
        return b.decode('hex')

    def write_to_stdout(data):
        sys.stdout.write(data + '\n')
        sys.stdout.flush()

    # The standard zlib.compressobj() accepts only positional arguments.
    def zlib_compressobj(level=6, method=zlib.DEFLATED, wbits=15, memlevel=8,
                         strategy=zlib.Z_DEFAULT_STRATEGY):
        return zlib.compressobj(level, method, wbits, memlevel, strategy)

    unicode = unicode
    bytes = str

elif is_py3:
    from urllib.parse import urlencode, urlparse, urlsplit

    imap = map

    def to_byte(char):
        return char

    def decode_hex(b):
        return bytes.fromhex(b)

    def write_to_stdout(data):
        sys.stdout.buffer.write(data + b'\n')
        sys.stdout.buffer.flush()

    zlib_compressobj = zlib.compressobj

    if is_py3_3:
        ssl = ssl_compat
    else:
        import ssl

    unicode = str
    bytes = bytes



from rs4 import asyncore
import socket

class WhoisRequest(asyncore.dispatcher_with_send):
    # simple whois requestor

    def __init__(self, consumer, query, host, port=43):
        asyncore.dispatcher_with_send.__init__(self)

        self.consumer = consumer
        self.query = query

        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.connect((host, port))

    def handle_connect(self):
        self.send(self.query.encode ("utf8") + b"\r\n")

    def handle_expt(self):
        self.close() # connection failed, shutdown
        self.consumer.abort()

    def handle_read(self):
        # get data from server
        self.consumer.feed(self.recv(2048))

    def handle_close(self):
        self.close()
        self.consumer.close()

class WhoisConsumer:

    def __init__(self, host):
        self.text = ""
        self.host = host

    def feed(self, text):
        self.text = self.text + text

    def abort(self):
        print(self.host, "=>", "failed")

    def close(self):
        print(self.host, "=>")
        print(self.text)

#
# try it out
for host in []:
    consumer = WhoisConsumer(host)
    request = WhoisRequest(consumer, host, "whois.internic.net")

# loop returns when all requests have been processed
asyncore.loop()



import threading
from rs4 import asyncore

class thread_safe_socket_map (dict):
	def __init__ (self):
		self.lock = threading.RLock ()

	def __setitem__ (self, k, v):
		with self.lock:
			dict.__setitem__ (self, k, v)

	def __getitem__ (self, k):
		with self.lock:
			v = dict.__getitem__ (self, k)
		return v

	def __delitem__ (self, k):
		with self.lock:
			dict.__delitem__ (self, k)

	def __contains__ (self, k):
		with self.lock:
			v = dict.__contains__ (self, k)
		return v

	def __len__ (self):
		with self.lock:
			v = dict.__len__ (self)
		return v

	def __nonzero__(self):
		with self.lock:
			v = dict.__len__ (self)
		return v

	def has_key (self, k):
		with self.lock:
			v = dict.has_key (self, k)
		return v

	def get (self, k, d = None):
		with self.lock:
			v = dict.get (self, k, d)
		return v

	def set (self, k, v):
		with self.lock:
			dict.set (self, k, v)

	def popitem (self, k):
		with self.lock:
			v = dict.popitem (self, k)
		return v

	def items (self):
		with self.lock:
			v = list (dict.items (self))
		return v

	def keys (self):
		with self.lock:
			v = list (dict.keys (self))
		return v

	def values (self):
		with self.lock:
			v = list (dict.values (self))
		return v


if not hasattr (asyncore, "_socket_map"):
	asyncore._socket_map = asyncore.socket_map
	del asyncore.socket_map
	asyncore.socket_map = thread_safe_socket_map ()



from .synsqlite3 import SynConnect
from .dbconnect import DBConnect

class Postgres (SynConnect):
    def connect (self):
        try:
            host, port = self.address
            self.conn = psycopg2.connect (
                dbname = self.dbname,
                user = self.user,
                password = self.password,
                host = host,
                port = port
            )
        except:
            self.handle_error ()
        else:
            self.connected = True

    def close_if_over_keep_live (self):
        DBConnect.close_if_over_keep_live (self)

    def execute (self, request, *args, **kargs):
        DBConnect.begin_tran (self, request)
        sql = self._compile (request)

        if not self.connected:
            self.connect ()
            self.conn.isolation_level = None

        try:
            if self.cur is None:
                self.cur = self.conn.cursor ()
            sql = self._compile (request)
            self.cur.execute (sql, *request.params [1:])
            self.has_result = True
        except:
            self.handle_error ()
        else:
            self.end_tran ()

class Redis (Postgres):
    def close (self, deactive = 1):
        if self.conn:
            self.conn.close ()
            self.conn = None
        self.connected = False
        DBConnect.close (self, deactive)

    def end_tran (self):
        if self.request:
            self.request.handle_result (None, self.expt, self.fetchall ())
            self.request = None
        super ().end_tran ()

    def connect (self):
        host, port = self.address
        self.conn = redis.Redis (host = host, port = port, db = self.dbname)

    def fetchall (self):
        result, self.response = self.response, None
        return result

    def prefetch (self):
        self.response = [[]]
        self.response = getattr (self.conn, self.request.method) (*self.request.params)
        self.has_result = True

    def execute (self, request, *args, **kargs):
        DBConnect.begin_tran (self, request)
        try:
            if not self.connected:
                self.connect ()
            self.prefetch ()
        except:
            self.handle_error ()
        else:
            self.end_tran ()

class MongoDB (Redis):
    def prefetch (self):
        self.response = None
        self.response = getattr (self.conn [self.request.dbname], self.request.method.lower ()) (*self.request.params)
        self.has_result = True

    def connect (self):
        user, password = "", ""
        auth = self.request.auth
        if auth:
            if len (auth) == 2:
                user, password = auth
            else:
                user = auth [0]
        host, port = self.address

        kargs = {}
        if user: kargs ["username"] = user
        if password: kargs ["password"] = password
        if port: kargs ["port"] = port
        self.conn = pymongo.MongoClient (host = host, **kargs)


from ..uninstalled import Uninstalled

try:
    import psycopg2
except ImportError:
    Postgres = Uninstalled ('psycopg2')

try:
    import redis
except ImportError:
    Redis = Uninstalled ('redis')

try:
    import pymongo
except ImportError:
    MongoDB = Uninstalled ('pymongo')



# -*- coding: utf-8 -*-
"""
hyper/http20/connection
~~~~~~~~~~~~~~~~~~~~~~~

Objects that build hyper's connection-level HTTP/2 abstraction.
"""
from .h2 import connection as h2_connection
from .h2 import events as h2_events
from .h2 import settings as h2_settings

from ..compat import ssl
from ..tls import wrap_socket, H2_NPN_PROTOCOLS, H2C_PROTOCOL
from ..common.exceptions import ConnectionResetError
from ..common.bufsocket import BufferedSocket
from ..common.headers import HTTPHeaderMap
from ..common.util import to_host_port_tuple, to_native_string, to_bytestring
from ..compat import unicode, bytes
from .stream import Stream
from .response import HTTP20Response, HTTP20Push
from .window import FlowControlManager
from .exceptions import ConnectionError, StreamResetError
from . import errors

import errno
import logging
import socket
import time
import threading

log = logging.getLogger(__name__)

DEFAULT_WINDOW_SIZE = 65535

TRANSIENT_SSL_ERRORS = (ssl.SSL_ERROR_WANT_READ, ssl.SSL_ERROR_WANT_WRITE)


class _LockedObject(object):
    """
    A wrapper class that hides a specific object behind a lock.

    The goal here is to provide a simple way to protect access to an object
    that cannot safely be simultaneously accessed from multiple threads. The
    intended use of this class is simple: take hold of it with a context
    manager, which returns the protected object.
    """
    def __init__(self, obj):
        self.lock = threading.RLock()
        self._obj = obj

    def __enter__(self):
        self.lock.acquire()
        return self._obj

    def __exit__(self, _exc_type, _exc_val, _exc_tb):
        self.lock.release()


class HTTP20Connection(object):
    """
    An object representing a single HTTP/2 connection to a server.

    This object behaves similarly to the Python standard library's
    ``HTTPConnection`` object, with a few critical differences.

    Most of the standard library's arguments to the constructor are irrelevant
    for HTTP/2 or not supported by hyper.

    :param host: The host to connect to. This may be an IP address or a
        hostname, and optionally may include a port: for example,
        ``'http2bin.org'``, ``'http2bin.org:443'`` or ``'127.0.0.1'``.
    :param port: (optional) The port to connect to. If not provided and one
        also isn't provided in the ``host`` parameter, defaults to 443.
    :param secure: (optional) Whether the request should use TLS. Defaults to
        ``False`` for most requests, but to ``True`` for any request issued to
        port 443.
    :param window_manager: (optional) The class to use to manage flow control
        windows. This needs to be a subclass of the
        :class:`BaseFlowControlManager
        <hyper.http20.window.BaseFlowControlManager>`. If not provided,
        :class:`FlowControlManager <hyper.http20.window.FlowControlManager>`
        will be used.
    :param enable_push: (optional) Whether the server is allowed to push
        resources to the client (see
        :meth:`get_pushes() <hyper.HTTP20Connection.get_pushes>`).
    :param ssl_context: (optional) A class with custom certificate settings.
        If not provided then hyper's default ``SSLContext`` is used instead.
    :param proxy_host: (optional) The proxy to connect to.  This can be an IP
        address or a host name and may include a port.
    :param proxy_port: (optional) The proxy port to connect to. If not provided
        and one also isn't provided in the ``proxy`` parameter, defaults to
        8080.
    """
    def __init__(self, host, port=None, secure=None, window_manager=None,
                 enable_push=False, ssl_context=None, proxy_host=None,
                 proxy_port=None, force_proto=None, **kwargs):
        """
        Creates an HTTP/2 connection to a specific server.
        """
        if port is None:
            self.host, self.port = to_host_port_tuple(host, default_port=443)
        else:
            self.host, self.port = host, port

        if secure is not None:
            self.secure = secure
        elif self.port == 443:
            self.secure = True
        else:
            self.secure = False

        self._enable_push = enable_push
        self.ssl_context = ssl_context

        # Setup proxy details if applicable.
        if proxy_host:
            if proxy_port is None:
                self.proxy_host, self.proxy_port = to_host_port_tuple(
                    proxy_host, default_port=8080
                )
            else:
                self.proxy_host, self.proxy_port = proxy_host, proxy_port
        else:
            self.proxy_host = None
            self.proxy_port = None

        #: The size of the in-memory buffer used to store data from the
        #: network. This is used as a performance optimisation. Increase buffer
        #: size to improve performance: decrease it to conserve memory.
        #: Defaults to 64kB.
        self.network_buffer_size = 65536

        self.force_proto = force_proto

        # Concurrency
        #
        # Use one lock (_lock) to synchronize any interaction with global
        # connection state, e.g. stream creation/deletion.
        #
        # It's ok to use the same in lock all these cases as they occur at
        # different/linked points in the connection's lifecycle.
        #
        # Use another 2 locks (_write_lock, _read_lock) to synchronize
        # - _send_cb
        # - _recv_cb
        # respectively.
        #
        # I.e, send/recieve on the connection and its streams are serialized
        # separately across the threads accessing the connection.  This is a
        # simple way of providing thread-safety.
        #
        # _write_lock and _read_lock synchronize all interactions between
        # streams and the connnection.  There is a third I/O callback,
        # _close_stream, passed to a stream's constructor.  It does not need to
        # be synchronized, it uses _send_cb internally (which is serialized);
        # its other activity (safe deletion of the stream from self.streams)
        # does not require synchronization.
        #
        # _read_lock may be acquired when already holding the _write_lock,
        # when they both held it is always by acquiring _write_lock first.
        #
        # Either _read_lock or _write_lock may be acquired whilst holding _lock
        # which should always be acquired before either of the other two.
        self._lock = threading.RLock()
        self._write_lock = threading.RLock()
        self._read_lock = threading.RLock()

        # Create the mutable state.
        self.__wm_class = window_manager or FlowControlManager
        self.__init_state()

        return

    def __init_state(self):
        """
        Initializes the 'mutable state' portions of the HTTP/2 connection
        object.

        This method exists to enable HTTP20Connection objects to be reused if
        they're closed, by resetting the connection object to its basic state
        whenever it ends up closed. Any situation that needs to recreate the
        connection can call this method and it will be done.

        This is one of the only methods in hyper that is truly private, as
        users should be strongly discouraged from messing about with connection
        objects themselves.
        """
        self._conn = _LockedObject(h2_connection.H2Connection())

        # Streams are stored in a dictionary keyed off their stream IDs. We
        # also save the most recent one for easy access without having to walk
        # the dictionary.
        #
        # We add a set of all streams that we or the remote party forcefully
        # closed with RST_STREAM, to avoid encountering issues where frames
        # were already in flight before the RST was processed.
        #
        # Finally, we add a set of streams that recently received data.  When
        # using multiple threads, this avoids reading on threads that have just
        # acquired the I/O lock whose streams have already had their data read
        # for them by prior threads.
        self.streams = {}
        self.recent_stream = None
        self.next_stream_id = 1
        self.reset_streams = set()
        self.recent_recv_streams = set()

        # The socket used to send data.
        self._sock = None

        # Instantiate a window manager.
        self.window_manager = self.__wm_class(65535)

        return

    def ping(self, opaque_data):
        """
        Send a PING frame.

        Concurrency
        -----------

        This method is thread-safe.

        :param opaque_data: A bytestring of length 8 that will be sent in the
                            PING frame.
        :returns: Nothing
        """
        self.connect()
        with self._write_lock:
            with self._conn as conn:
                conn.ping(to_bytestring(opaque_data))
            self._send_outstanding_data()

    def request(self, method, url, body=None, headers=None):
        """
        This will send a request to the server using the HTTP request method
        ``method`` and the selector ``url``. If the ``body`` argument is
        present, it should be string or bytes object of data to send after the
        headers are finished. Strings are encoded as UTF-8. To use other
        encodings, pass a bytes object. The Content-Length header is set to the
        length of the body field.

        Concurrency
        -----------

        This method is thread-safe.

        :param method: The request method, e.g. ``'GET'``.
        :param url: The URL to contact, e.g. ``'/path/segment'``.
        :param body: (optional) The request body to send. Must be a bytestring
            or a file-like object.
        :param headers: (optional) The headers to send on the request.
        :returns: A stream ID for the request.
        """
        headers = headers or {}

        # Concurrency
        #
        # It's necessary to hold a lock while this method runs to satisfy H2
        # protocol requirements.
        #
        # - putrequest obtains the next valid new stream_id
        # - endheaders sends a http2 message using the new stream_id
        #
        # If threads interleave these operations, it could result in messages
        # being sent in the wrong order, which can lead to the out-of-order
        # messages with lower stream IDs being closed prematurely.
        with self._write_lock:
            stream_id = self.putrequest(method, url)

            default_headers = (':method', ':scheme', ':authority', ':path')
            for name, value in headers.items():
                is_default = to_native_string(name) in default_headers
                self.putheader(name, value, stream_id, replace=is_default)

            # Convert the body to bytes if needed.
            if body and isinstance(body, (unicode, bytes)):
                body = to_bytestring(body)

            self.endheaders(message_body=body, final=True, stream_id=stream_id)

            return stream_id

    def _get_stream(self, stream_id):
        if stream_id is None:
            return self.recent_stream
        elif stream_id in self.reset_streams or stream_id not in self.streams:
            raise StreamResetError("Stream forcefully closed")
        else:
            return self.streams[stream_id]

    def get_response(self, stream_id=None):
        """
        Should be called after a request is sent to get a response from the
        server. If sending multiple parallel requests, pass the stream ID of
        the request whose response you want. Returns a
        :class:`HTTP20Response <hyper.HTTP20Response>` instance.
        If you pass no ``stream_id``, you will receive the oldest
        :class:`HTTPResponse <hyper.HTTP20Response>` still outstanding.

        Concurrency
        -----------

        This method is thread-safe.

        :param stream_id: (optional) The stream ID of the request for which to
            get a response.
        :returns: A :class:`HTTP20Response <hyper.HTTP20Response>` object.
        """
        stream = self._get_stream(stream_id)
        return HTTP20Response(stream.getheaders(), stream)

    def get_pushes(self, stream_id=None, capture_all=False):
        """
        Returns a generator that yields push promises from the server. **Note
        that this method is not idempotent**: promises returned in one call
        will not be returned in subsequent calls. Iterating through generators
        returned by multiple calls to this method simultaneously results in
        undefined behavior.

        :param stream_id: (optional) The stream ID of the request for which to
            get push promises.
        :param capture_all: (optional) If ``False``, the generator will yield
            all buffered push promises without blocking. If ``True``, the
            generator will first yield all buffered push promises, then yield
            additional ones as they arrive, and terminate when the original
            stream closes.
        :returns: A generator of :class:`HTTP20Push <hyper.HTTP20Push>` objects
            corresponding to the streams pushed by the server.
        """
        stream = self._get_stream(stream_id)
        for promised_stream_id, headers in stream.get_pushes(capture_all):
            yield HTTP20Push(
                HTTPHeaderMap(headers), self.streams[promised_stream_id]
            )

    def connect(self):
        """
        Connect to the server specified when the object was created. This is a
        no-op if we're already connected.

        Concurrency
        -----------

        This method is thread-safe. It may be called from multiple threads, and
        is a noop for all threads apart from the first.

        :returns: Nothing.

        """
        with self._lock:
            if self._sock is not None:
                return

            if not self.proxy_host:
                host = self.host
                port = self.port
            else:
                host = self.proxy_host
                port = self.proxy_port

            sock = socket.create_connection((host, port))

            if self.secure:
                assert not self.proxy_host, "Proxy with HTTPS not supported."
                sock, proto = wrap_socket(sock, host, self.ssl_context,
                                          force_proto=self.force_proto)
            else:
                proto = H2C_PROTOCOL

            log.debug("Selected NPN protocol: %s", proto)
            assert proto in H2_NPN_PROTOCOLS or proto == H2C_PROTOCOL

            self._sock = BufferedSocket(sock, self.network_buffer_size)

            self._send_preamble()

    def _connect_upgrade(self, sock):
        """
        Called by the generic HTTP connection when we're being upgraded. Locks
        in a new socket and places the backing state machine into an upgrade
        state, then sends the preamble.
        """
        self._sock = sock

        with self._conn as conn:
            conn.initiate_upgrade_connection()
            conn.update_settings(
                {h2_settings.ENABLE_PUSH: int(self._enable_push)}
            )
        self._send_outstanding_data()

        # The server will also send an initial settings frame, so get it.
        # However, we need to make sure our stream state is set up properly
        # first, or any extra data we receive might cause us problems.
        s = self._new_stream(local_closed=True)
        self.recent_stream = s

        self._recv_cb()

    def _send_preamble(self):
        """
        Sends the necessary HTTP/2 preamble.
        """
        # We need to send the connection header immediately on this
        # connection, followed by an initial settings frame.
        with self._conn as conn:
            conn.initiate_connection()
            conn.update_settings(
                {h2_settings.ENABLE_PUSH: int(self._enable_push)}
            )
        self._send_outstanding_data()

        # The server will also send an initial settings frame, so get it.
        self._recv_cb()

    def close(self, error_code=None):
        """
        Close the connection to the server.

        Concurrency
        -----------

        This method is thread-safe.

        :param error_code: (optional) The error code to reset all streams with.
        :returns: Nothing.
        """
        # Concurrency
        #
        # It's necessary to hold the lock here to ensure that threads closing
        # the connection see consistent state, and to prevent creation of
        # of new streams while the connection is being closed.
        #
        # I/O occurs while the lock is held; waiting threads will see a delay.
        with self._lock:
            # Close all streams
            for stream in list(self.streams.values()):
                log.debug("Close stream %d" % stream.stream_id)
                stream.close(error_code)

            # Send GoAway frame to the server
            try:
                with self._conn as conn:
                    conn.close_connection(error_code or 0)
                self._send_outstanding_data(tolerate_peer_gone=True)
            except Exception as e:  # pragma: no cover
                log.warn("GoAway frame could not be sent: %s" % e)

            if self._sock is not None:
                self._sock.close()
            self.__init_state()

    def _send_outstanding_data(self, tolerate_peer_gone=False,
                               send_empty=True):
        # Concurrency
        #
        # Hold _write_lock; getting and writing data from _conn is synchronized
        #
        # I/O occurs while the lock is held; waiting threads will see a delay.
        with self._write_lock:
            with self._conn as conn:
                data = conn.data_to_send()
            if data or send_empty:
                self._send_cb(data, tolerate_peer_gone=tolerate_peer_gone)

    def putrequest(self, method, selector, **kwargs):
        """
        This should be the first call for sending a given HTTP request to a
        server. It returns a stream ID for the given connection that should be
        passed to all subsequent request building calls.

        Concurrency
        -----------

        This method is thread-safe. It can be called from multiple threads,
        and each thread should receive a unique stream ID.

        :param method: The request method, e.g. ``'GET'``.
        :param selector: The path selector.
        :returns: A stream ID for the request.
        """
        # Create a new stream.
        s = self._new_stream()

        # To this stream we need to immediately add a few headers that are
        # HTTP/2 specific. These are: ":method", ":scheme", ":authority" and
        # ":path". We can set all of these now.
        s.add_header(":method", method)
        s.add_header(":scheme", "https" if self.secure else "http")
        s.add_header(":authority", self.host)
        s.add_header(":path", selector)

        # Save the stream.
        self.recent_stream = s

        return s.stream_id

    def putheader(self, header, argument, stream_id=None, replace=False):
        """
        Sends an HTTP header to the server, with name ``header`` and value
        ``argument``.

        Unlike the ``httplib`` version of this function, this version does not
        actually send anything when called. Instead, it queues the headers up
        to be sent when you call
        :meth:`endheaders() <hyper.HTTP20Connection.endheaders>`.

        This method ensures that headers conform to the HTTP/2 specification.
        In particular, it strips out the ``Connection`` header, as that header
        is no longer valid in HTTP/2. This is to make it easy to write code
        that runs correctly in both HTTP/1.1 and HTTP/2.

        :param header: The name of the header.
        :param argument: The value of the header.
        :param stream_id: (optional) The stream ID of the request to add the
            header to.
        :returns: Nothing.
        """
        stream = self._get_stream(stream_id)
        stream.add_header(header, argument, replace)

        return

    def endheaders(self, message_body=None, final=False, stream_id=None):
        """
        Sends the prepared headers to the server. If the ``message_body``
        argument is provided it will also be sent to the server as the body of
        the request, and the stream will immediately be closed. If the
        ``final`` argument is set to True, the stream will also immediately
        be closed: otherwise, the stream will be left open and subsequent calls
        to ``send()`` will be required.

        :param message_body: (optional) The body to send. May not be provided
            assuming that ``send()`` will be called.
        :param final: (optional) If the ``message_body`` parameter is provided,
            should be set to ``True`` if no further data will be provided via
            calls to :meth:`send() <hyper.HTTP20Connection.send>`.
        :param stream_id: (optional) The stream ID of the request to finish
            sending the headers on.
        :returns: Nothing.
        """
        self.connect()

        stream = self._get_stream(stream_id)

        headers_only = (message_body is None and final)

        # Concurrency:
        #
        # Hold _write_lock: synchronize access to the connection's HPACK
        # encoder and decoder and the subsquent write to the connection
        with self._write_lock:
            stream.send_headers(headers_only)

            # Send whatever data we have.
            if message_body is not None:
                stream.send_data(message_body, final)

            self._send_outstanding_data()

        return

    def send(self, data, final=False, stream_id=None):
        """
        Sends some data to the server. This data will be sent immediately
        (excluding the normal HTTP/2 flow control rules). If this is the last
        data that will be sent as part of this request, the ``final`` argument
        should be set to ``True``. This will cause the stream to be closed.

        :param data: The data to send.
        :param final: (optional) Whether this is the last bit of data to be
            sent on this request.
        :param stream_id: (optional) The stream ID of the request to send the
            data on.
        :returns: Nothing.
        """
        stream = self._get_stream(stream_id)
        stream.send_data(data, final)

        return

    def _new_stream(self, stream_id=None, local_closed=False):
        """
        Returns a new stream object for this connection.
        """
        # Concurrency
        #
        # Hold _lock: ensure that threads accessing the connection see
        # self.next_stream_id in a consistent state
        #
        # No I/O occurs, the delay in waiting threads depends on their number.
        with self._lock:
            s = Stream(
                stream_id or self.next_stream_id,
                self.__wm_class(DEFAULT_WINDOW_SIZE),
                self._conn,
                self._send_outstanding_data,
                self._recv_cb,
                self._stream_close_cb,
            )
            s.local_closed = local_closed
            self.streams[s.stream_id] = s
            self.next_stream_id += 2

            return s

    def _send_cb(self, data, tolerate_peer_gone=False):
        """
        This is the callback used by streams to send data on the connection.

        This acts as a dumb wrapper around the socket send method.
        """
        # Concurrency
        #
        # Hold _write_lock: ensures only writer at a time
        #
        # I/O occurs while the lock is held; waiting threads will see a delay.
        with self._write_lock:
            try:
                self._sock.sendall(data)
            except socket.error as e:
                if (not tolerate_peer_gone or
                        e.errno not in (errno.EPIPE, errno.ECONNRESET)):
                    raise

    def _adjust_receive_window(self, frame_len):
        """
        Adjusts the window size in response to receiving a DATA frame of length
        ``frame_len``. May send a WINDOWUPDATE frame if necessary.
        """
        # Concurrency
        #
        # Hold _write_lock; synchronize the window manager update and the
        # subsequent potential write to the connection
        #
        # I/O may occur while the lock is held; waiting threads may see a
        # delay.
        with self._write_lock:
            increment = self.window_manager._handle_frame(frame_len)

            if increment:
                with self._conn as conn:
                    conn.increment_flow_control_window(increment)
                self._send_outstanding_data(tolerate_peer_gone=True)

        return

    def _single_read(self):
        """
        Performs a single read from the socket and hands the data off to the
        h2 connection object.
        """
        # Begin by reading what we can from the socket.
        #
        # Concurrency
        #
        # Synchronizes reading the data
        #
        # I/O occurs while the lock is held; waiting threads will see a delay.
        with self._read_lock:
            if self._sock is None:
                raise ConnectionError('tried to read after connection close')
            self._sock.fill()
            data = self._sock.buffer.tobytes()
            self._sock.advance_buffer(len(data))
            with self._conn as conn:
                events = conn.receive_data(data)
            stream_ids = set(getattr(e, 'stream_id', -1) for e in events)
            stream_ids.discard(-1)  # sentinel
            stream_ids.discard(0)  # connection events
            self.recent_recv_streams |= stream_ids

        for event in events:
            if self.recent_stream and self.recent_stream in self.streams:
                self.streams [self.recent_stream].events.append (events)
            if isinstance(event, h2_events.DataReceived):
                self._adjust_receive_window(event.flow_controlled_length)
                self.streams[event.stream_id].receive_data(event)
            elif isinstance(event, h2_events.PushedStreamReceived):
                if self._enable_push:
                    self._new_stream(event.pushed_stream_id, local_closed=True)
                    self.streams[event.parent_stream_id].receive_push(event)
                else:
                    # Servers are forbidden from sending push promises when
                    # the ENABLE_PUSH setting is 0, but the spec leaves the
                    # client action undefined when they do it anyway. So we
                    # just refuse the stream and go about our business.
                    self._send_rst_frame(event.pushed_stream_id, 7)
            elif isinstance(event, h2_events.ResponseReceived):
                self.streams[event.stream_id].receive_response(event)
            elif isinstance(event, h2_events.TrailersReceived):
                self.streams[event.stream_id].receive_trailers(event)
            elif isinstance(event, h2_events.StreamEnded):
                self.streams[event.stream_id].receive_end_stream(event)
            elif isinstance(event, h2_events.StreamReset):
                if event.stream_id not in self.reset_streams:
                    self.reset_streams.add(event.stream_id)
                    self.streams[event.stream_id].receive_reset(event)
            elif isinstance(event, h2_events.ConnectionTerminated):
                # If we get GoAway with error code zero, we are doing a
                # graceful shutdown and all is well. Otherwise, throw an
                # exception.
                self.close()

                # If an error occured, try to read the error description from
                # code registry otherwise use the frame's additional data.
                if event.error_code != 0:
                    try:
                        name, number, description = errors.get_data(
                            event.error_code
                        )
                    except ValueError:
                        error_string = (
                            "Encountered error code %d" % event.error_code
                        )
                    else:
                        error_string = (
                            "Encountered error %s %s: %s" %
                            (name, number, description)
                        )

                    raise ConnectionError(error_string)
            else:
                log.info("Received unhandled event %s", event)

        self._send_outstanding_data(tolerate_peer_gone=True, send_empty=False)

    def _recv_cb(self, stream_id=0):
        """
        This is the callback used by streams to read data from the connection.

        This stream reads what data it can, and throws it into the underlying
        connection, before farming out any events that fire to the relevant
        streams. If the socket remains readable, it will then optimistically
        continue to attempt to read.

        This is generally called by a stream, not by the connection itself, and
        it's likely that streams will read a frame that doesn't belong to them.

        :param stream_id: (optional) The stream ID of the stream reading data
            from the connection.

        """
        # Begin by reading what we can from the socket.
        #
        # Concurrency
        #
        # Ignore this read if some other thread has recently read data from
        # from the requested stream.
        #
        # The lock here looks broad, but is needed to ensure correct behavior
        # when there are multiple readers of the same stream.  It is
        # re-acquired in the calls to self._single_read.
        #
        # I/O occurs while the lock is held; waiting threads will see a delay.
        with self._read_lock:
            log.debug('recv for stream %d with %s already present',
                      stream_id,
                      self.recent_recv_streams)
            if stream_id in self.recent_recv_streams:
                self.recent_recv_streams.discard(stream_id)
                return

            # make sure to validate the stream is readable.
            # if the connection was reset, this stream id won't appear in
            # self.streams and will cause this call to raise an exception.
            if stream_id:
                self._get_stream(stream_id)

            # TODO: Re-evaluate this.
            self._single_read()
            count = 9
            retry_wait = 0.05  # can improve responsiveness to delay the retry

            while count and self._sock is not None and self._sock.can_read:
                # If the connection has been closed, bail out, but retry
                # on transient errors.
                try:
                    self._single_read()
                except ConnectionResetError:
                    break
                except ssl.SSLError as e:  # pragma: no cover
                    # these are transient errors that can occur while reading
                    # from ssl connections.
                    if e.args[0] in TRANSIENT_SSL_ERRORS:
                        continue
                    else:
                        raise
                except socket.error as e:  # pragma: no cover
                    if e.errno in (errno.EINTR, errno.EAGAIN):
                        # if 'interrupted' or 'try again', continue
                        time.sleep(retry_wait)
                        continue
                    elif e.errno == errno.ECONNRESET:
                        break
                    else:
                        raise

                count -= 1

    def _send_rst_frame(self, stream_id, error_code):
        """
        Send reset stream frame with error code and remove stream from map.
        """
        # Concurrency
        #
        # Hold _write_lock; synchronize generating the reset frame and writing
        # it
        #
        # I/O occurs while the lock is held; waiting threads will see a delay.
        with self._write_lock:
            with self._conn as conn:
                conn.reset_stream(stream_id, error_code=error_code)
            self._send_outstanding_data()

        # Concurrency
        #
        # Hold _lock; the stream storage is being updated. No I/O occurs, any
        # delay is proportional to the number of waiting threads.
        with self._lock:
            try:
                del self.streams[stream_id]
                self.recent_recv_streams.discard(stream_id)
            except KeyError as e:  # pragma: no cover
                log.warn(
                    "Stream with id %d does not exist: %s",
                    stream_id, e)

            # Keep track of the fact that we reset this stream in case there
            # are other frames in flight.
            self.reset_streams.add(stream_id)

    def _stream_close_cb(self, stream_id):
        """
        Called by a stream when it is closing, so that state can be cleared.
        """
        try:
            del self.streams[stream_id]
            self.recent_recv_streams.discard(stream_id)
        except KeyError:
            pass

    # The following two methods are the implementation of the context manager
    # protocol.
    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()
        return False  # Never swallow exceptions.



from rs4 import asynchat, asyncore
import re, os, sys
import ssl
import socket
import time
import zlib
from warnings import warn
from errno import ECONNRESET, ENOTCONN, ESHUTDOWN, ECONNABORTED, EWOULDBLOCK
if os.name == "nt":
    from errno import WSAENOTCONN
import select
import threading
from . import adns
from ..protocols.http2 import H2_PROTOCOLS
from ..athreads.fifo import await_fifo
from rs4.ssl_ import resolve_cert_reqs, resolve_ssl_version, create_urllib3_context
from collections import deque
from ..protocols.http import respcodes
from .baseconnect import ConnectProxy, BaseConnect, DEFAULT_ZOMBIE_TIMEOUT, DEFAULT_KEEP_ALIVE
import random

DEBUG = False

class SocketPanic (Exception): pass
class TimeOut (Exception): pass

class ConnectProxy (ConnectProxy):
    def execute_late (self):
        if self._canceled ():
            return
        request, rs, handler, handle_request_failed = self._args
        self._asyncon.set_timeout (self._timeout)
        try:
            self._asyncon.execute (request, rs, handler)
        except:
            handle_request_failed (rs, self._asyncon)
        self._args = None


class AsynConnect (asynchat.async_chat, BaseConnect):
    ac_in_buffer_size = 65535
    ac_out_buffer_size = 65535
    fifo_class = deque
    keep_connect = True
    is_ssl = False

    def __init__ (self, address, lock, logger = None):
        BaseConnect.__init__ (self, lock, logger)

        self.address = address
        self.handler = None

        self.auth = None
        self.proxy = False
        self.initialize_connection ()
        self._closed = False
        self.backend = False

        self.ac_in_buffer = b''
        self.incoming = []
        self.producer_fifo = self.fifo_class ()
        self.set_event_time ()
        self.__sendlock = threading.Lock ()

        asyncore.dispatcher.__init__(self)

    def __repr__ (self):
        return "<AsynConnect %s:%d>" % self.address

    def execute (self, request, rs, handler, *args, **kargs):
        r = handler (self, request, rs.handle_result)
        if self.get_proto () and self.isconnected ():
            self.handler.handle_request (r)
        else:
            r.handle_request ()

    def duplicate (self):
        new_asyncon = self.__class__ (self.address, self.lock, self.logger)
        # used in socketpool
        new_asyncon.proxy = self.proxy
        # used in skitai cluster manager
        new_asyncon.keep_alive = self.keep_alive
        new_asyncon.auth = self.auth
        new_asyncon.backend = self.backend
        new_asyncon.keep_connect = self.keep_connect
        return new_asyncon

    def set_auth (self, auth):
        self.auth = auth

    def get_auth (self):
        return self.auth

    def close (self):
        # sometimes socket is not closed at once
        # possibly related with SSL socket
        # then prevent doble callbacking in request_handler

        if self.socket:
            # self.socket is still None, when DNS not found
            asynchat.async_chat.close (self)
            self._fileno = None

        # re-init asychat
        self.ac_in_buffer = b''
        self.incoming = []
        self.producer_fifo.clear()
        self._proto = None
        self._closed = True

        if not self.handler:
            # return to the pool
            return self.set_active (False)

        if not self.errcode:
            # disconnect intentionally
            return

        handler, self.handller = self.handler, None
        keep_active = False
        try:
            keep_active = handler.connection_closed (self.errcode, self.errmsg)
        except:
            self.trace ()

        if not keep_active:
            self.set_active (False)
            if self.errcode not in (704, 712, 722):
                # controlled shutdown
                self.logger (
                    ".....socket %s has been closed (reason: %d)" % ("%s:%d" % self.address, self.errcode),
                    "info"
                )
        # DO NOT Change any props, because may be request has been restarted

    def end_tran (self):
        if DEBUG:
            self.logger ('end_tran {rid:%s} %s' % (self.handler.request.meta.get ('req_id', -1), self.handler.request.uri), 'debug')
        if not self.backend:
            self.del_channel ()
        self.set_active (False)
        if not self.keep_connect:
            self.disconnect ()

    def get_proto (self):
        with self.lock:
            p = self._proto
        return p

    def set_proto (self, proto):
        with self.lock:
            self._proto = proto

    def initialize_connection (self):
        self._raised_ENOTCONN = 0 # for win32
        self._history = []
        self._proto = None
        self._handshaking = False
        self._handshaked = False

        self.established = False
        self.upgraded = False

    def is_proxy (self):
        return self.proxy

    def log (self, msg, logtype):
        if self.handler is not None and hasattr (self.handler, "log"):
            self.handler.log (msg, logtype)
        elif self.logger:
            self.logger (msg, logtype)
        else:
            warn ("No logger")

    def trace (self):
        if self.handler is not None and hasattr (self.handler, "trace"):
            self.handler.trace ()
        elif self.logger:
            self.logger.trace ()
        else:
            warn ("No logger for traceback")

    def clean_shutdown_control (self, phase, time_in_this_phase):
        self._no_more_request = True
        if self.isactive () or (self.handler and self.handler.working ()):
            return 1
        else:
            self.handle_close (712, "Controlled Shutdown")
            self._no_more_request = False
            return 0

    def is_channel_in_map (self, map = None):
        if map is None:
            map = self._map
        return self._fileno in map

    def isconnected (self):
        with self.lock:
            r = self.connected
        return r

    def del_channel (self, map = None):
        fd = self._fileno
        if map is None:
            map = self._map
        if fd in map:
            del map[fd]

    def create_socket (self, family, type):
        self.family_and_type = family, type
        sock = socket.socket (family, type)
        sock.setblocking (0)
        self.set_socket (sock)

    def connect (self):
        if adns.query:
            adns.query (self.address [0], "A", callback = self.continue_connect)
        else:
            self.continue_connect ()

    def continue_connect (self, answer = None):
        self.initialize_connection ()
        if not adns.query:
            self.create_socket (socket.AF_INET, socket.SOCK_STREAM)
            try: asynchat.async_chat.connect (self, self.address)
            except:    self.handle_error (714)
            return

        ipaddr = answer [-1]["data"]
        if not ipaddr:
            return self.handle_close (704)
        else:
            port = self.address [1]

        self.create_socket (socket.AF_INET, socket.SOCK_STREAM)
        try:
            asynchat.async_chat.connect (self, (ipaddr, port))
        except:
            self.handle_error (714)

    def recv (self, buffer_size):
        try:
            data = self.socket.recv (buffer_size)
            if not data:
                self.handle_close (700, "Connection closed unexpectedly in recv")
                return b''
            else:
                self.set_event_time ()
                return data
        except socket.error as why:
            if why.errno in asyncore._DISCONNECTED:
                self.handle_close (700, "Connection closed unexpectedly in recv")
                return b''
            else:
                raise

    def send (self, data):
        try:
            numsent = self.socket.send (data)
            if numsent:
                self.set_event_time ()
            return numsent

        except socket.error as why:
            if why.errno == EWOULDBLOCK:
                return 0
            elif why.errno in asyncore._DISCONNECTED:
                if os.name == "nt" and why.errno == asyncore.ENOTCONN:
                    # winsock sometimes raise ENOTCONN and sometimes recovered.
                    # Found this error at http://file.hungryboarder.com:8080/HBAdManager/pa.html?siteId=hboarder&zoneId=S-2
                    if self._raised_ENOTCONN <= 3:
                        self._raised_ENOTCONN += 1
                        return 0
                    else:
                        self._raised_ENOTCONN = 0
                self.handle_close (700, "Connection closed unexpectedly in send")
                return 0
            else:
                raise

    def close_if_over_keep_live (self):
        if time.time () - self.event_time > self.keep_alive:
            self.disconnect ()

    def set_keep_alive (self, keep_alive = 10):
        self.keep_alive = keep_alive

    def handle_connect (self):
        if hasattr (self.handler, "has_been_connected"):
            self.handler.has_been_connected ()

    def handle_expt (self):
        self.handle_close (703)

    def handle_error (self, code = 701):
        self.trace ()
        self.handle_close (code)

    def handle_timeout (self):
        self.handle_close (702)

    def handle_expt_event (self):
        err = self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
        if err != 0:
            self.handle_close (703, "Socket %d Error" % err)
        else:
            self.handle_expt ()

    def maintern (self, object_timeout):
        if time.time () - self.event_time > object_timeout:
            if self.handler:
                if hasattr (self.handler, "control_shutdown"):
                    self.handler.control_shutdown ()
                self.handle_close (722)
            else:
                self.disconnect ()
            return True
        return False

    def writable (self):
        with self.__sendlock:
            return self.producer_fifo or (not self.connected)

    def handle_write (self):
        with self.__sendlock:
            self.initiate_send ()

    # proxy POST need no init_send
    def push (self, thing, init_send = True):
        if self.connected:
            self.close_if_over_keep_live () # check keep-alive

        if isinstance(thing, (bytes, bytearray, memoryview)):
            with self.__sendlock:
                asynchat.async_chat.push (self, thing)
        else:
            self.push_with_producer (thing, init_send)

    def push_with_producer (self, producer, init_send = True):
        if self.connected:
            self.close_if_over_keep_live () # check keep-alive

        with self.__sendlock:
            self.producer_fifo.append (producer)
            if init_send:
                self.initiate_send ()

    def handle_abort (self):
        self.close ()

    def handle_close (self, code = 700, msg = ""):
        if code == 0: msg = ""
        self.errcode = code
        if msg:
            self.errmsg = msg
        else:
            self.errmsg = respcodes.get (code, "Undefined Error")
        self.close ()

    def collect_incoming_data (self, data):
        if not self.handler:
            if self.connected:
                self.logger ("recv data but no handler, droping data %d" % len (data), "warn")
                self.disconnect ()
            return
        self.handler.collect_incoming_data (data)

    def found_terminator (self):
        if not self.handler:
            if self.connected:
                self.logger ("found terminator but no handler", "warn")
                self.disconnect ()
            return # already closed
        self.handler.found_terminator ()

    def disconnect (self):
        # no error
        self.handle_close (0)

    def reconnect (self):
        self.disconnect ()
        self.connect ()

    def set_proxy (self, flag = True):
        self.proxy = flag

    def begin_tran (self, handler):
        if self._no_more_request:
            return self.handle_close (705)
        # IMP: the reason why DNS error, _closed must be located here
        self._closed = False
        self.errcode = 0
        self.errmsg = ""

        self.handler = handler
        if DEBUG:
            self.logger ('begin_tran {rid:%s} %s' % (self.handler.request.meta.get ('req_id', -1), self.handler.request.uri), 'debug')

        self.set_event_time ()
        try:
            if not self.connected:
                self.connect ()
            elif not self.backend:
                self.add_channel ()
        except:
            self.handle_error ()


class AsynSSLConnect (AsynConnect):
    is_ssl = True

    def __init__ (self, address, lock = None, logger = None):
        AsynConnect.__init__ (self, address, lock, logger)
        self.ac_negotiate_http2 = True

    def negotiate_http2 (self, flag):
        self.ac_negotiate_http2 = flag

    def handshake (self):
        if not self._handshaking:
            err = self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
            if err != 0:
                raise OSError(err, asyncore._strerror(err))

            ssl_context = create_urllib3_context(ssl_version=resolve_ssl_version(None), cert_reqs=resolve_cert_reqs(None))
            if self.ac_negotiate_http2:
                try: ssl_context.set_alpn_protocols (H2_PROTOCOLS)
                except AttributeError: ssl_context.set_npn_protocols (H2_PROTOCOLS)
            self.socket = ssl_context.wrap_socket (self.socket, do_handshake_on_connect = False, server_hostname = self.address [0])
            self._handshaking = True

        try:
            self.socket.do_handshake ()
        except ssl.SSLError as why:
            if why.args [0] in (ssl.SSL_ERROR_WANT_READ, ssl.SSL_ERROR_WANT_WRITE):
                return False
            raise ssl.SSLError(why)

        try: self._proto = self.socket.selected_alpn_protocol()
        except (AttributeError, NotImplementedError):
            try: self._proto = self.socket.selected_npn_protocol()
            except (AttributeError, NotImplementedError): pass

        self._handshaked = True
        return True

    def handle_connect_event (self):
        try:
            if not self._handshaked and not self.handshake ():
                return
        except:
            return self.handle_error (713)
        AsynConnect.handle_connect_event (self)

    def recv (self, buffer_size):
        if self._closed:
            # usually handshaking failure, already handled exception
            return b''

        try:
            data = self.socket.recv (buffer_size)
            if not data:
                self.handle_close (700, "Connection closed unexpectedly")
                return b''
            else:
                self.set_event_time ()
                return data

        except ssl.SSLError as why:
            if why.errno == ssl.SSL_ERROR_WANT_READ:
                try:
                    raise BlockingIOError
                except NameError:
                    raise socket.error (EWOULDBLOCK)

            # closed connection
            elif why.errno in (ssl.SSL_ERROR_ZERO_RETURN, ssl.SSL_ERROR_EOF):
                self.handle_close (700, "Connection closed by SSL_ERROR_ZERO_RETURN or SSL_ERROR_EOF")
                return b''

            else:
                raise

    def send (self, data):
        if self._closed:
            # usually handshaking failure, already handled exception
            return

        try:
            numsent = self.socket.send (data)
            if numsent:
                self.set_event_time ()
            return numsent

        except ssl.SSLError as why:
            if why.errno == ssl.SSL_ERROR_WANT_WRITE:
                return 0
            elif why.errno == ssl.SSL_ERROR_ZERO_RETURN:
                self.handle_close (700, "Connection closed by SSL_ERROR_ZERO_RETURN")
                return 0
            else:
                raise


class AsynSSLProxyConnect (AsynSSLConnect, AsynConnect):
    is_ssl = True

    def __init__ (self, address, lock = None, logger = None):
        AsynConnect.__init__ (self, address, lock, logger)

    def handle_connect_event (self):
        if self.established:
            AsynSSLConnect.handle_connect_event (self)
        else:
            AsynConnect.handle_connect_event (self)

    def recv (self, buffer_size):
        if self._handshaked or self._handshaking:
            return AsynSSLConnect.recv (self, buffer_size)
        else:
            return AsynConnect.recv (self, buffer_size)

    def send (self, data):
        if self._handshaked or self._handshaking:
            return AsynSSLConnect.send (self, data)
        else:
            return AsynConnect.send (self, data)




import sys, time
from rs4 import asyncore
import gc
import select
import os
import bisect
import socket
import time
import types
from .protocols import dns
from collections import deque
import threading

try:
	from pympler import muppy, summary, tracker
except ImportError:
	pass

if os.name == "nt":
	from errno import WSAENOTSOCK

_shutdown_phase = 0
_shutdown_timeout = 30 # seconds per phase
_exit_code = 0
_last_maintern = 0
_maintern_interval = 3.0
_killed_zombies = 0
_select_errors = 0
_poll_count = 0
_polling = 0
_logger = None

EXHAUST_DNS = True

class TickTimer:
	def __init__ (self):
		self.onces = []
		self._canceled = {}
		self._call_id = 0
		self._lock = threading.Lock ()

	def later (self, delay, func, args = ()):
		return self.at (time.monotonic () + delay, func, args)

	def at (self, at, func, args = ()):
		# at is time.monotonic ()
		with self._lock:
			self._call_id += 1
			call_id = self._call_id
			last_at = self.onces and self.onces [-1][1] or None
			self.onces.append ((call_id, at, func, args))
			if last_at and at < last_at:
				self.onces.sort (key = lambda x: x [1])
		return call_id

	def cancel (self, call_id):
		with self._lock:
			self._canceled [call_id] = None

	def tick (self):
		with self._lock:
			if not self.onces:
				return
		now = time.monotonic ()
		funcs = []
		with self._lock:
			for i in range (len (self.onces)):
				at = self.onces [0][1]
				if at > now:
					break
				call_id, at, func, args = self.onces.pop (0)
				if call_id in self._canceled:
					self._canceled.pop (call_id)
					continue
				funcs.append ((func, args))
		for func, args in funcs:
			func (*args)


class Maintern:
	def __init__ (self):
		self.q = []

	# mainternacing -----------------------------------------
	def sched (self, interval, func, args = None):
		now = time.time ()
		self.q.append ((now + interval, interval, func, args))
		self.q.sort (key = lambda x: x [0])
		#bisect.insort (self.q, (now + interval, interval, func, args))

	def __call__ (self, now):
		excutes = 0
		for exetime, interval, func, args in self.q:
			if exetime > now: break
			excutes += 1
			if args:
				func (now, *args)
			else:
				func (now)

		for i in range (excutes):
			exetime, interval, func, args = self.q.pop (0)
			#bisect.insort (self.q, (now + interval, interval, func, args))
			self.q.append ((now + interval, interval, func, args))
			self.q.sort (key = lambda x: x [0])

def maintern_gc (now):
	gc.collect ()

def maintern_zombie_channel (now, map = None):
	global _killed_zombies

	if map is None:
		map = asyncore.socket_map

	for channel in map.values():
		if hasattr (channel, "handle_timeout"):
			try:
				# +3 is make gap between server & client
				iszombie = (now - channel.event_time) > channel.zombie_timeout
			except AttributeError:
				continue
			if iszombie:
				_killed_zombies += 1
				try:
					channel.handle_timeout ()
				except:
					channel.handle_error ()

maintern = None
tick_timer = None
def init (kill_zombie_interval = 10.0, logger = None):
	global tick_timer, maintern, _logger

	_logger = logger
	maintern = Maintern ()
	maintern.sched (kill_zombie_interval, maintern_zombie_channel)
	maintern.sched (300.0, maintern_gc)
	tick_timer = TickTimer ()

summary_tracker = None
def summary_track (now):
	global summary_tracker

	all_objects = muppy.get_objects ()
	#all_objects = muppy.filter (all_objects, Type = dict)
	#for each in all_objects:
	#	print (each)
	sum1 = summary.summarize (all_objects)
	summary.print_ (sum1)
	print ('-' * 79)
	if summary_tracker is None:
		summary_tracker = tracker.SummaryTracker ()
	summary_tracker.print_diff ()

def enable_memory_track (interval = 10.0):
	global maintern
	maintern.sched (interval, summary_track)

def shutdown (exit_code, shutdown_timeout = 30.0):
	global _shutdown_phase
	global _shutdown_timeout
	global _exit_code

	if _shutdown_phase:
		# aready entered
		return

	if _shutdown_phase == 0:
		_exit_code = exit_code
		_shutdown_phase = 1

	_shutdown_timeout = shutdown_timeout

def loop (timeout = 30.0):
	global _shutdown_phase
	global _shutdown_timeout
	global _exit_code
	global _polling
	global maintern

	if maintern is None:
		init ()

	_shutdown_phase = 0
	_shutdown_timeout = 30
	_exit_code = 0
	_polling = 1

	try:
		lifetime_loop(timeout)
	except KeyboardInterrupt:
		graceful_shutdown_loop()
	else:
		graceful_shutdown_loop()

	_polling = 0

def remove_notsocks (map):
	global _select_errors, _logger

	# on Windows we can get WSAENOTSOCK if the client
	# rapidly connect and disconnects
	killed = 0
	for fd, obj in map.items():
		r = []; w = []; e = []
		is_r = obj.readable()
		is_w = obj.writable()
		if is_r:
			r = [fd]
		# accepting sockets should not be writable
		if is_w and not obj.accepting:
			w = [fd]
		if is_r or is_w:
			e = [fd]

		try:
			select.select (r, w, e, 0)

		except:
			#_logger and _logger.trace ()
			killed += 1
			_select_errors += 1

			try:
				try: obj.handle_expt ()
				except: obj.handle_error ()
			except:
				_logger and _logger.trace ()

			try: del map [fd]
			except KeyError: pass

	return killed

poll_fun = asyncore.poll

def poll_fun_wrap (timeout, map = None):
	global _logger, poll_fun

	if map is None:
		map = asyncore.socket_map

	if EXHAUST_DNS:
		dns.pop_all ()

	try:
		poll_fun (timeout, map)

	except (SystemError, KeyboardInterrupt):
		raise

	except (TypeError, OSError) as why:
		_logger and _logger.trace ()
		killed = remove_notsocks (map) # WSAENOTSOCK
		if not killed:
			# no errors, restart worker
			raise

	except ValueError:
		_logger and _logger.trace ()
		# negative file descriptor, testing all sockets
		# or too many file descriptors in select(), divide and conquer
		if not remove_notsocks (map):
			half = int (len (map) / 2)
			tmap = {}
			cc = 0
			for k, v in map.items ():
				tmap [k] = v
				cc += 1
				if cc == half:
					poll_fun_wrap (timeout, tmap)
					tmap = {}
			poll_fun_wrap (timeout, tmap)

	except:
		_logger and _logger.trace ()
		raise


def lifetime_loop (timeout = 30.0, count = 0):
	global _last_maintern
	global _maintern_interval
	global tick_timer

	loop = 0
	map = asyncore.socket_map

	while map and _shutdown_phase == 0:
		poll_fun_wrap (timeout, map)
		tick_timer.tick ()
		now = time.time ()
		if (now - _last_maintern) > _maintern_interval:
			maintern (now)
			_last_maintern = now
		loop += 1
		if count and loop > count:
			break

def graceful_shutdown_loop ():
	global _shutdown_phase
	global tick_timer

	timestamp = time.time()
	timeout = 1.0
	map = asyncore.socket_map
	while map and _shutdown_phase < 4:
		time_in_this_phase = time.time() - timestamp
		veto = 0
		for fd,obj in map.items():
			try:
				fn = getattr (obj,'clean_shutdown_control')
			except AttributeError:
				pass
			else:
				try:
					veto = veto or fn (_shutdown_phase, time_in_this_phase)
				except:
					obj.handle_error()

		if veto and time_in_this_phase < _shutdown_timeout:
			poll_fun_wrap (timeout, map)
			tick_timer.tick ()
		else:
			_shutdown_phase += 1
			timestamp = time.time()



from ..protocols import dns
import time
import threading
import os

class DNSCache:
	maintern_interval = 17
	too_many_errors = 3
	def __init__ (self, prefer_protocol, logger = None):
		self.logger = logger
		self.prefer_protocol = prefer_protocol
		self.__last_maintern = time.time ()
		self.__lock = threading.RLock ()
		self.cache = {"localhost": {"data": "127.0.0.1", "valid": 2**32, "typename": "A"}}
		self.errors = {}
		self.hits = 0

	def set (self, answers):
		if not answers [-1]["data"]:
			name = answers [0]["name"]
			if name not in self.errors:
				self.errors [name] = 1
			else:
				self.errors [name] += 1
			return

		with self.__lock:
			for answer in answers:
				name = answer['name'].lower ()
				try:
					ttl = int (answer ["ttl"])
				except (KeyError, ValueError):
					ttl = 1800
				answer ["valid"] = time.time () + max (ttl, 1800)
				if name not in self.cache:
					self.cache [name] = answer

	def expire (self, host):
		with self.__lock:
			try: del self.cache [host]
			except KeyError: pass

	def maintern (self, now):
		deletables = []
		with self.__lock:
			for host in self.cache:
				ans = self.cache [host]
				if ans ["valid"] < now:
					deletables.append (host)
			for host in deletables:
				del self.cache [host]
		self.__last_maintern = time.time ()

	def get (self, host, qtype = "A", check_ttl = True):
		now = time.time ()
		if now - self.__last_maintern > self.maintern_interval:
			self.maintern (now)

		host = host.lower ()
		with self.__lock:
			while 1:
				try:
					answer = self.cache [host]
				except KeyError:
					if self.errors.get (host, 0) >= self.too_many_errors:
						return [{"name": host, "data": None, "error": "too many error"} ]
					return []
				else:
					tn = answer ['typename']
					if tn == "CNAME":
						host = answer ['data'].lower ()
					elif tn == qtype:
						break

		if check_ttl and answer ["valid"] < now:
			# extend 30 seconds for other querees
			answer ['valid'] = now + (answer ["data"] and 30 or 1)
			# new query will be started
			return []

		else:
			return [answer]

	def is_ip (self, name):
		arr = name.split (".")
		if len (arr) != 4: return False
		try: arr = [x for x in map (int, arr) if x & 255 == x]
		except ValueError:
			return False
		if len (arr) != 4: return False
		return True

	def __call__ (self, host, qtype, callback):
		self.hits += 1
		hit = self.get (host, qtype, True)
		if hit:
			return callback (hit)

		if self.is_ip (host):
			self.set ([{"name": host, "data": host, "typename": qtype}])
			return callback ([{"name": host, "data": host, "typename": qtype}])

		try:
			dns.query (host, qtype = qtype, protocol = self.prefer_protocol, callback = [self.set, callback])
		except:
			self.logger.trace (host)
			hit = [{"name": host, "data": None, "typename": qtype, 'ttl': 60}]
			self.set (hit)
			callback (hit)


query = None

def init (logger, dns_servers = [], prefer_protocol = "tcp"):
	dns.create_pool (dns_servers, logger)
	global query
	if query is None:
		query = DNSCache (prefer_protocol, logger)

def get (name, qtype = "A"):
	global query
	return query.get (name, qtype, False)

